% SOLVE  solves PDE on cartesian mesh with piecewise constant
% material, p and use of static condenstation
%
% function [u, q, K] = solve(x, y, p, AE, c, f);
%
% x   x-coordinates of nodes of the mesh
% y   y-coordinates of nodes of the mesh
% p   polynomial degree or polynomial degree per cell
% AE  constant AE or AE per cell
% c   constant c or c per cell
% f   function handle of source term
%
% u   solution vector
% E   energy norm
%
function [u, E] = solve(x, y, p, AE, c, f);

% intervall lengthes
hx = diff(x); nx = length(hx);
hy = diff(y); ny = length(hy);

if nx < 2 | ny < 2
  disp('At least 2 intervals needed. But only one given.');
  K = []; b = [];
  return;
end

% number of degrees of freedom
N = (nx*p-1)*(ny*p-1) - nx*ny*(p-1)^2;
q = zeros (N,1);
K = [];

%% ** Assembling **
tic,
for j=1:ny
  for i=1:nx
%    fprintf('i = %d, j = %d\n', i, j);

    % element matrix and element load vector
    [Kloc, qloc] = elemMatrixVector(p, AE, c, f, x(i), y(j), hx(i), hy(j));

    % global indices I (without interior), local indices I0, I1
    % mapping: I0(j) -> I(j)  for all j
    [I, I0, I1] = indices(p, i, j, nx, ny);
    
    % static condensation
    % Equation (3.2.33), (3.2.34) on p. 68 of Schwab's hp-FEM book
    Kq(I0,[I0 max(I0)+1]) = ...
	[Kloc(I0,I0) qloc(I0)];  % length(I0) x (length(I0)+1)
    if p > 1
      Kq(I0,[I0 max(I0)+1]) = Kq(I0,[I0 max(I0)+1]) ...
	  - Kloc(I0,I1)*full(Kloc(I1,I1) \ [Kloc(I1,I0) qloc(I1)]);
    end
    
    % cancel indices on the boundary
    % mapping: I0(j) -> I(j)  for all j    
    [I, I0] = cancelDiriBC(I, I0, p, i, j, nx, ny);    
            
    % adding in global matrix and vector of rhs
    K(I,I) = K(I,I) + Kq(I0,I0);
	 
    % adding in vector of rhs
    q(I)   = q(I)   + Kq(I0,end);
  end
end
toc

%% ** Solving the system of equations **
tic,u = K \ q;toc
size(u);

E = 0;

%% ** Computing the internal amplitudes **
tic
for j=1:ny
  for i=1:nx
    %fprintf('i = %d, j = %d\n', i, j);

    % element matrix and element load vector
    [Kloc, qloc] = elemMatrixVector(p, AE, c, f, x(i), y(j), hx(i), hy(j));
    % global indices I, local indices I0, I1
    [I, I0, I1] = indices(p, i, j, nx, ny);
    % cancel indices on the boundary
    [I, I0] = cancelDiriBC(I, I0, p, i, j, nx, ny);
    if p > 1    
      uloc = Kloc(I1,I1) \ (qloc(I1) - Kloc(I1,I0)*u(I));
      %uloc = zeros(size(I1))';
      u = [u; uloc];
    else
      uloc = [];
    end
    uloc = [u(I);uloc];
    
    E = E + uloc' * qloc([I0 I1]);
  end
end
toc


return;

% ELEMMATRIXVECTOR  returns local matrix and rhs vector 
%
% function [Kloc, qloc, Iloc] = elemMatrixVector(p, AE, c, f, xi, yj, hx, hy);
%
% input
%   p     polynomial degree
%   AE    formula for AE(x)
%   c     formula for c(x)
%   xi    left x-coordinate of the cell
%   yj    lower y-coordinate of the cell
%   hx    size in x-direction
%   hy    size in y-direction
%
% output
%   Kloc  local matrix
%   qloc  local rhs vector
%   Iloc  local indices

function [Kloc, qloc, Iloc] = elemMatrixVector(p, AE, c, f, xi, yj, hx, hy);

% computation of the matrices just once
persistent S M
if isempty(S) | (p+1)^2 ~= length(S)
  % reference element matrices in 1D
  M1D = refMass (p);
  S1D = refStiff(p);
  S = ...
  M = ...
end

% all local indices (with all edge and vertex dof)
Iloc = 1:(p+1)^2;

% Determinant of the Jacobian of the mapping Q from reference square
detJ = ...

% midpoint of the cell
xmean = xi+hx/2; ymean = yj+hy/2;
AEmean = AE(xmean,ymean);
cmean = c(xmean,ymean);

% element matrix
Kloc = ...

% 1D quadrature points
[xhat,w] = gauleg(p+10); yhat = xhat;
% shape functions on quadrature points
phi    = shapefct1D(p, xhat); 
F      = f(xi + hx*(xhat+1)/2, yj + hy*(yhat+1)/2);

qloc = ...

return;


% INDICES   returns the indices in one cell
%
% function [I, I0, I1] = indices(p, i, j, nx, ny)
%
% input
%   p     polynomial degree
%   i     horizontal index of the cell
%   j     vertical index of the cell
%   nx    number of cells in horizontal direction
%   ny    number of cells in vertical direction
%
% output
%   I     global indices (vertex and edge dof)
%   I0    belonging local indices
%   I1    local indices in the interior of the cell
%
function [I, I0, I1] = indices(p, i, j, nx, ny)

% local indices on vertices and edges
I0   = [1, 2, p+2, p+3 ...                           % vertices
	3:p+1, p+4:2*p+2, ...                        % horizontal edges
	2*p+3:p+1:p*(p+1)+1, 2*p+4:p+1:p*(p+1)+2];   % vertical edges

% interior local indices 
I1   = setdiff(1:(p+1)^2, I0);

Nnodes = (nx-1)*(ny-1);   % number of nodes
Nedhor = nx*(ny-1)*(p-1); % number of dof on horizontal edges
Nedver = (nx-1)*ny*(p-1); % number of dof on vertical edges

% global vertex indices
Inodes = ...
% global indices of horizontal edges
Iedhor = ((j-2)*nx + i - 1)*(p-1) + [1:p-1];
Iedhor = [Iedhor Iedhor+nx*(p-1)] + Nnodes;
% global indices of vertical edges
Iedver = ...
Iedver = ... + Nnodes + Nedhor;

% global indices (without interior)
I = [Inodes Iedhor Iedver];

return;

% CANCELDIRIBC    cancels the indices on the boundary
%
% function [I, I0] = cancelDiriBC(I, I0, p, i, j, nx, ny);
%
% input
%   I     global indices
%   I0    belonging local indices
%   p     polynomial degree
%   i     horizontal index of the cell
%   j     vertical index of the cell
%   nx    number of cells in horizontal direction
%   ny    number of cells in vertical direction
%
% output
%   I     global indices without those on the boundary
%   I0    belonging local indices (without those on the boundary)
%
function [I, I0] = cancelDiriBC(I, I0, p, i, j, nx, ny);

% cancel some local indices due to the Dirichlet boundary condition
Icancel = [];
if i == 1
  Icancel = [1, p+2, 2*p+3:p+1:p*(p+1)+1];  % cancel p+1 dof on left side
elseif i == nx
  Icancel = [2, p+3, 2*p+4:p+1:p*(p+1)+2];  % cancel p+1 dof on right side
end
if j == 1
  Icancel = [Icancel 1:p+1];                % cancel p+1 dof on bottom side
elseif j == ny
  Icancel = [Icancel p+2,p+3,p+4:2*p+2];    % cancel p+1 dof on top side
end
Iremain = find(~ismember(I0, Icancel));
I  = I (Iremain);
I0 = I0(Iremain);

return;

